﻿using System;
using System.IO;
using System.Windows.Media.Imaging;

namespace WpfGeoTagger.Models
{
	/// <summary>
	/// GeoTag を管理する為のクラスです。
	/// </summary>
	class GeoTag
	{
		/// <summary>
		/// 画像ファイルから GeoTag インスタンスを生成します。
		/// </summary>
		/// <param name="fileName">画像ファイルのパス。</param>
		/// <returns>成功時は GeoTag インスタンス。失敗時は null 参照。</returns>
		public static GeoTag FromImage( string fileName )
		{
			using( var stream = new FileStream( fileName, FileMode.Open, FileAccess.Read, FileShare.Read ) )
			{
				var decoder   = new JpegBitmapDecoder( stream, BitmapCreateOptions.None, BitmapCacheOption.None );
				var metadata  = ( BitmapMetadata )decoder.Frames[ 0 ].Metadata;

				// 緯度
				double latitude = 0.0;
				{
					var gpsLatitude = metadata.GetQuery( GeoTag.ExifPathGpsLatitude ) as ulong[];
					if( gpsLatitude == null ) { return null; }

					var direction = metadata.GetQuery( GeoTag.ExifPathGpsLatitudeDirection ) as string;
					if( direction == null ) { return null; }

					// 南半球なら負の値
					latitude = GeoTag.ConvertToGeoTagLocation( gpsLatitude );
					if( direction == "S" )
					{
						latitude *= -1;
					}
				}

				// 経度
				double longitude = 0.0;
				{
					var gpsLongitude = metadata.GetQuery( GeoTag.ExifPathGpsLongitude ) as ulong[];
					if( gpsLongitude == null ) { return null; }

					var direction = metadata.GetQuery( GeoTag.ExifPathGpsLongitudeDirection ) as string;
					if( direction == null ) { return null; }

					// 西半球なら負の値
					longitude = GeoTag.ConvertToGeoTagLocation( gpsLongitude );
					if( direction == "W" )
					{
						longitude *= -1;
					}
				}

				return new GeoTag() { Latitude = latitude, Longitude = longitude };
			}
		}

		/// <summary>
		/// EXIF から得られた GPS 情報を表す { 度, 分, 秒 } の配列から GeoTag 用の位置情報を取得します。 
		/// </summary>
		/// <param name="gpsValues">GPS 情報の配列。</param>
		/// <returns>GeoTag 用の位置情報。</returns>
		private static double ConvertToGeoTagLocation( ulong[] gpsValues )
		{
			double degrees = GeoTag.NormalizeGpsValue( gpsValues[ 0 ] );
			double minutes = GeoTag.NormalizeGpsValue( gpsValues[ 1 ] );
			double seconds = GeoTag.NormalizeGpsValue( gpsValues[ 2 ] );

			// GeoTag の位置へ変換
			return degrees + ( minutes / 60.0 ) + ( seconds / 3600 );
		}

		/// <summary>
		/// GeoTag の位置情報から EXIF 用の GPS 情報を取得します。
		/// </summary>
		/// <param name="location">GeoTag の位置情報。</param>
		/// <returns>EXIF 用の GPS 情報。</returns>
		private static ulong[] ConvertToGpsValues( double location )
		{
			location = Math.Abs( location );

			// 度
			double degrees = Math.Floor( location );
			location -= degrees;

			// 分
			double minutes = Math.Floor( location * 60.0 );
			location -= ( minutes / 60.0 );

			// 秒
			double seconds = Math.Round( location * 3600.0 * 1000000.0 );

			// GPS 情報へ変換する
			ulong[] values = new ulong[ 3 ];
			values[ 0 ] = Convert.ToUInt64( degrees + 0x100000000 );
			values[ 1 ] = Convert.ToUInt64( minutes + 0x100000000 );
			values[ 2 ] = GeoTag.ConvertToGpsSeconds( seconds );

			return values;
		}

		/// <summary>
		/// GeoTag の秒データから GPS 用の秒データを取得します。
		/// </summary>
		/// <param name="seconds">GeoTag の秒。</param>
		/// <returns>作成した GPS 用の秒データ。</returns>
		private static ulong ConvertToGpsSeconds( double seconds )
		{
			byte[] num = BitConverter.GetBytes( ( int )seconds );
			byte[] den = BitConverter.GetBytes( ( int )1000000 );
			byte[] sec = new byte[ 8 ] { num[ 0 ], num[ 1 ], num[ 2 ], num[ 3 ], den[ 0 ], den[ 1 ], den[ 2 ], den[ 3 ] };

			return BitConverter.ToUInt64( sec, 0 );
		}

		/// <summary>
		/// GPS 情報の一つのデータ ( GPSLatitude または GPSLongitude ) を正規化します。
		/// EXIF のタグ情報については、以下のページを参照して下さい。
		/// <span>
		/// http://homepage1.nifty.com/gigo/DC/GPS/gpsifd.html
		/// </span>
		/// </summary>
		/// <param name="value">GPS 情報。</param>
		/// <returns>正規化されたデータ。</returns>
		private static double NormalizeGpsValue( ulong value )
		{
			byte[]	bytes = BitConverter.GetBytes( value );
			int		upper = BitConverter.ToInt32( bytes, 0 );
			int		lower = BitConverter.ToInt32( bytes, 4 );

			return ( ( double )upper / ( double )lower );
		}

		/// <summary>
		/// 現在の GeoTag を EXIF の GPS 情報としてファイルに保存します。
		/// </summary>
		/// <param name="fileName">ファイルへのパス。</param>
		public void Save( string fileName )
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// 現在の GeoTag を表す System.String を返します。
		/// </summary>
		/// <returns>GeoTag の内容を表す System.String。</returns>
		public override string ToString()
		{
			return String.Format( "緯度 = {0}、経度 = {1}", this.Latitude, this.Longitude );
		}

		/// <summary>
		/// 緯度を取得または設定します。
		/// </summary>
		public double Latitude { get; set; }

		/// <summary>
		/// 経度を取得または設定します。
		/// </summary>
		public double Longitude { get; set; }

		#region 定数

		/// <summary>
		/// EXIF の GPS 情報における、緯度を示すパス文字列。
		/// </summary>
		private const string ExifPathGpsLatitude = "/app1/ifd/gps/subifd:{ulong=2}";

		/// <summary>
		/// EXIF の GPS 情報における、緯度の進行方向 ( "N" または "S" ) を示すパス文字列。
		/// </summary>
		private const string ExifPathGpsLatitudeDirection = "/app1/ifd/gps/subifd:{short=1}";

		/// <summary>
		/// EXIF の GPS 情報における、緯度を示すパス文字列。
		/// </summary>
		private const string ExifPathGpsLongitude = "/app1/ifd/gps/subifd:{ulong=4}";

		/// <summary>
		/// EXIF の GPS 情報における、経度の進行方向 ( "E" または "W" ) を示すパス文字列。
		/// </summary>
		private const string ExifPathGpsLongitudeDirection = "/app1/ifd/gps/subifd:{short=3}";

		#endregion
	}
}
