﻿using System;
using System.ComponentModel;
using System.IO;
using System.Windows.Media.Imaging;
using WpfGeoTagger.Models;

namespace WpfGeoTagger.ViewModels
{
	/// <summary>
	/// Image の Model と View を仲介します。
	/// </summary>
	class ImageViewModel : ViewModelBase
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="id">画像の識別子。</param>
		/// <param name="fileName">画像ファイルのパス。</param>
		/// <param name="onGeoTagLoadedd">GeoTag の読み込みが完了した事を通知するイベント ハンドラ。</param>
		/// <exception cref="FileNotFoundException">画像ファイルが存在しません。</exception>
		public ImageViewModel( int id, string fileName, Action< ImageViewModel > onGeoTagLoadedd )
		{
			if( !File.Exists( fileName ) ) { throw new FileNotFoundException( fileName ); }

			this.Id   = id;
			this.Name = Path.GetFileName( fileName );
			
			this.LoadImage( fileName );
			this.LoadGeoTag( fileName, onGeoTagLoadedd );
		}

		/// <summary>
		/// 画像から Geo タグ情報を読み込みます。
		/// </summary>
		/// <param name="fileName">画像ファイルのパス。</param>
		private void LoadGeoTag( string fileName, Action< ImageViewModel > onGeoTagLoadedd )
		{
			if( !File.Exists( fileName ) ) { return; }

			BackgroundWorker worker = new BackgroundWorker();
	
			worker.DoWork += ( sender, e ) =>
			{
				e.Result = GeoTag.FromImage( fileName );
			};
		
			worker.RunWorkerCompleted += ( sender, e ) =>
			{
				var g = e.Result as GeoTag;
				if( g != null )
				{
					this.GeoTag = g;
					this.NotifyPropertyChanged( "GeoTag" );

					if( onGeoTagLoadedd != null )
					{
						onGeoTagLoadedd( this );
					}
				}
			};

			worker.RunWorkerAsync();
		}

		/// <summary>
		/// 画像を読み込みます。
		/// </summary>
		/// <param name="fileName">画像ファイルのパス。</param>
		private void LoadImage( string fileName )
		{
			if( File.Exists( fileName ) )
			{
				BackgroundWorker worker = new BackgroundWorker();

				worker.DoWork += ( sender, e ) =>
				{
					BitmapImage image = new BitmapImage();
					image.BeginInit();
					image.DecodePixelWidth = 64;
					image.CacheOption = BitmapCacheOption.OnLoad;
					image.UriSource = new Uri( fileName );
					image.EndInit();
					image.Freeze();

					e.Result = image;
				};

				worker.RunWorkerCompleted += ( sender, e ) =>
				{
					var image = e.Result as BitmapImage;
					if( image != null )
					{
						this.Image = image;
						this.NotifyPropertyChanged( "Image" );
					}
				};

				worker.RunWorkerAsync();
			}
		}

		/// <summary>
		/// GeoTag を設定または取得します。
		/// </summary>
		public GeoTag GeoTag { get; set; }
		
		/// <summary>
		/// 識別子を取得します。
		/// </summary>
		public int Id { get; private set; }

		/// <summary>
		/// 画像を設定または取得します。
		/// </summary>
		public BitmapImage Image { get; private set; }

		/// <summary>
		/// 画像ファイル名を取得します。
		/// </summary>
		public string Name { get; private set; }
	}
}
